import sys

from .factory import SerializerFactory


def main():
    args = sys.argv

    if len(args) != 5:
        print("Invalid amount of args")
        exit()

    _, filepath_from, filepath_to, format_from, format_to = args

    # if format_from == format_to:
    #     print("Formats cant be equal")
    #     exit()

    try:
        source_serializer = SerializerFactory.create_serializer(format_from)
        result_serializer = SerializerFactory.create_serializer(format_to)
    except Exception as err:
        print(err)
        exit()

    try:
        with open(filepath_from) as file_from, open(filepath_to, "w") as file_to:
            deserialized_object = source_serializer.load(file_from)
            result_serializer.dump(deserialized_object, file_to)
    except Exception as err:
        print(err)
        exit()



import types
from enum import StrEnum, auto


class TYPE(StrEnum):
    BYTES = auto()
    FUNCTION = auto()
    CELL = auto()
    CLASS = auto()
    ITERATOR = auto()
    CODE = auto()
    OBJECT = auto()
    MODULE = auto()
    TUPLE = auto()
    SET = auto()


UNSERIALIZABLE_DUNDER = (
    "__mro__",
    "__base__",
    "__basicsize__",
    "__class__",
    "__dictoffset__",
    "__name__",
    "__qualname__",
    "__text_signature__",
    "__itemsize__",
    "__flags__",
    "__weakrefoffset__",
    "__objclass__",
)

UNSERIALIZABLE_TYPES = (
    types.WrapperDescriptorType,
    types.MethodDescriptorType,
    types.BuiltinFunctionType,
    types.MappingProxyType,
    types.GetSetDescriptorType,
)

UNSERIALIZABLE_CODE_TYPES = (
    "co_positions",
    "co_lines",
    "co_exceptiontable",
    "co_lnotab",
)



from abc import abstractmethod
from typing import IO


class BaseSerializer:
    @staticmethod
    @abstractmethod
    def dumps(obj) -> str:
        pass

    @staticmethod
    @abstractmethod
    def dump(obj, fp: IO[str]) -> None:
        pass

    @staticmethod
    @abstractmethod
    def loads(s: str) -> object:
        pass

    @staticmethod
    @abstractmethod
    def load(fp: IO[str]) -> object:
        pass



from ..serializers.base_serializer import BaseSerializer


class CustomSerializer:
    def __init__(self, encoder, serializer: BaseSerializer):
        self._encoder = encoder
        self._serializer = serializer

    def dumps(self, obj):
        return self._serializer.dumps(self._encoder.encode(obj))

    def dump(self, obj, fp):
        return self._serializer.dump(self._encoder.encode(obj), fp)

    def loads(self, s):
        return self._encoder.decode(self._serializer.loads(s))

    def load(self, fp):
        return self._encoder.decode(self._serializer.load(fp))



import base64
import builtins
import inspect
import types

from .constants import (
    TYPE,
    UNSERIALIZABLE_DUNDER,
    UNSERIALIZABLE_TYPES,
    UNSERIALIZABLE_CODE_TYPES,
)
from .helpers import is_iterable, get_class_that_defined_method

PRIMITIVES = (int, str, bool, float, types.NoneType)


class Encoder:
    @classmethod
    def encode(cls, obj):
        if isinstance(obj, PRIMITIVES):
            return obj

        if isinstance(obj, types.ModuleType):
            return cls._encode_module(obj)

        if isinstance(obj, types.CellType):
            return cls._encode_cell(obj)

        if isinstance(obj, bytes):
            return cls._encode_bytes(obj)

        if isinstance(obj, list):
            return type(obj)((cls.encode(item) for item in obj))

        if isinstance(obj, (tuple, set)):
            return cls._encode_collection(obj)

        if isinstance(obj, dict):
            return {key: cls.encode(value) for key, value in obj.items()}

        if isinstance(obj, (types.FunctionType, types.MethodType)):
            return cls._encode_function(obj)

        if isinstance(obj, type):
            return cls._encode_class(obj)

        if isinstance(obj, types.CodeType):
            return cls._encode_code(obj)

        if is_iterable(obj):
            return cls._encode_iterator(obj)

        if isinstance(obj, object):
            return cls._encode_object(obj)

    @classmethod
    def decode(cls, obj) -> any:
        if isinstance(obj, PRIMITIVES):
            return obj

        if isinstance(obj, list):
            return type(obj)((cls.decode(item) for item in obj))

        if isinstance(obj, dict):
            type_to_decode = cls._get_type(obj)

            if type_to_decode is None:
                return {key: cls.decode(value) for key, value in obj.items()}

            if type_to_decode == TYPE.BYTES:
                return cls._decode_bytes(obj)

            if type_to_decode == TYPE.FUNCTION:
                return cls._decode_function(obj)

            if type_to_decode == TYPE.CELL:
                return cls._decode_cell(obj)

            if type_to_decode == TYPE.CLASS:
                return cls._decode_class(obj)

            if type_to_decode == TYPE.ITERATOR:
                return cls._decode_iterator(obj)

            if type_to_decode == TYPE.CODE:
                return cls._decode_code(obj)

            if type_to_decode == TYPE.OBJECT:
                return cls._decode_object(obj)

            if type_to_decode == TYPE.MODULE:
                return cls._decode_module(obj)

            if type_to_decode in (TYPE.TUPLE, TYPE.SET):
                return cls._decode_collection(obj)

        return obj

    @classmethod
    def _encode_bytes(cls, bytes_obj: bytes):
        data = base64.b64encode(bytes_obj).decode("ascii")
        return cls._mark_data(data, TYPE.BYTES)

    @classmethod
    def _decode_bytes(cls, encoded):
        return base64.b64decode(cls._get_data(encoded).encode("ascii"))

    @classmethod
    def _encode_module(cls, obj):
        return cls._mark_data(obj.__name__, TYPE.MODULE)

    @classmethod
    def _decode_module(cls, encoded):
        return __import__(cls._get_data(encoded))

    @classmethod
    def _encode_collection(cls, obj):
        data = [cls.encode(item) for item in obj]
        return cls._mark_data(data, type(obj).__name__.lower())

    @classmethod
    def _decode_collection(cls, obj):
        data = cls._get_data(obj)
        collection = getattr(builtins, cls._get_type(obj).lower())
        return collection((cls.decode(item) for item in data))

    @classmethod
    def _encode_code(cls, code):
        attrs = [attr for attr in dir(code) if attr.startswith("co")]

        code_dict = {
            attr: cls.encode(getattr(code, attr))
            for attr in attrs
            if attr not in UNSERIALIZABLE_CODE_TYPES
        }

        return cls._mark_data(code_dict, TYPE.CODE)

    @classmethod
    def _decode_code(cls, obj):
        data = cls._get_data(obj)

        def f():
            pass

        code_dict = cls.decode(data)

        return f.__code__.replace(**code_dict)

    @classmethod
    def _encode_function(cls, func):
        fcode = func.__code__
        fname = func.__name__
        fdefaults = func.__defaults__
        fdict = func.__dict__
        fclass = get_class_that_defined_method(func)
        fclosure = (
            tuple(cell for cell in func.__closure__ if cell.cell_contents is not fclass)
            if func.__closure__ is not None
            else tuple()
        )

        fglobs = {
            key: cls.encode(value)
            for (key, value) in func.__globals__.items()
            if key in func.__code__.co_names
            and value is not fclass
            and key != func.__code__.co_name
        }

        encoded_function = cls.encode(
            dict(
                code=fcode,
                name=fname,
                argdefs=fdefaults,
                closure=fclosure,
                fdict=fdict,
                globals=fglobs,
            )
        )

        return cls._mark_data(
            encoded_function,
            TYPE.FUNCTION,
            is_method=isinstance(func, types.MethodType),
        )

    @classmethod
    def _decode_function(cls, encoded):
        encoded_function = cls.decode(cls._get_data(encoded))

        fdict = encoded_function.pop("fdict")

        new_func = types.FunctionType(**encoded_function)
        new_func.__dict__.update(fdict)
        new_func.__globals__.update({new_func.__name__: new_func})
        return new_func

    @classmethod
    def _encode_cell(cls, obj):
        data = cls.encode(obj.cell_contents)
        return cls._mark_data(data, TYPE.CELL)

    @classmethod
    def _decode_cell(cls, obj):
        return cls._create_cell(cls.decode(cls._get_data(obj)))

    @classmethod
    def _encode_class(cls, obj):
        data = {
            attr: cls.encode(getattr(obj, attr))
            for attr, value in inspect.getmembers(obj)
            if attr not in UNSERIALIZABLE_DUNDER
            and type(value) not in UNSERIALIZABLE_TYPES
        }

        data["__bases__"] = [
            cls.encode(base) for base in obj.__bases__ if base != object
        ]

        data["__name__"] = obj.__name__

        return cls._mark_data(data, TYPE.CLASS)

    @classmethod
    def _decode_class(cls, obj):
        data = cls._get_data(obj)

        class_bases = tuple(cls.decode(base) for base in data.pop("__bases__"))
        class_dict = {
            attr: cls.decode(value)
            for (attr, value) in data.items()
            if not (isinstance(value, dict) and cls._get_type(value) == TYPE.FUNCTION)
        }

        result = type(data["__name__"], class_bases, class_dict)

        for key, value in data.items():
            if isinstance(value, dict) and cls._get_type(value) == TYPE.FUNCTION:
                try:
                    func = cls.decode(value)
                except ValueError:
                    closure = cls._get_data(value)["closure"]
                    cls._get_data(closure).append(cls._create_cell(result))
                    func = cls.decode(value)

                func.__globals__.update({result.__name__: result})

                if value.get("is_method"):
                    func = types.MethodType(func, result)

                setattr(result, key, func)

        return result

    @classmethod
    def _encode_iterator(cls, obj):
        data = list(map(cls.encode, obj))
        return cls._mark_data(data, TYPE.ITERATOR)

    @classmethod
    def _decode_iterator(cls, obj):
        data = cls._get_data(obj)
        return iter(cls.decode(value) for value in data)

    @classmethod
    def _encode_object(cls, obj):
        data = {
            "__class__": cls.encode(obj.__class__),
            "attrs": {
                attr: cls.encode(value)
                for (attr, value) in inspect.getmembers(obj)
                if not attr.startswith("__")
                and not isinstance(value, types.FunctionType)
                and not isinstance(value, types.MethodType)
            },
        }

        return cls._mark_data(data, TYPE.OBJECT)

    @classmethod
    def _decode_object(cls, obj):
        data = cls._get_data(obj)
        obj_class = cls.decode(data["__class__"])

        result = object.__new__(obj_class)
        result.__dict__ = {
            key: cls.decode(value) for key, value in data["attrs"].items()
        }

        return result

    @staticmethod
    def _mark_data(data, _type: TYPE, **additional_props):
        return dict(__type=_type, data=data, **additional_props)

    @staticmethod
    def _get_type(encoded):
        if isinstance(encoded, dict):
            return encoded.get("__type")

    @staticmethod
    def _get_data(encoded):
        if isinstance(encoded, dict):
            return encoded.get("data")

    @staticmethod
    def _create_cell(value):
        return (lambda: value).__closure__[0]



from typing import IO

from .base_serializer import BaseSerializer


class JsonSerializer(BaseSerializer):
    @classmethod
    def dumps(cls, obj) -> str:
        if isinstance(obj, bool):
            return str(obj).lower()

        if isinstance(obj, (int, float)):
            return str(obj)

        if isinstance(obj, str):
            return f'"{obj}"'

        if isinstance(obj, type(None)):
            return "null"

        if isinstance(obj, (list, tuple)):
            return f"[{','.join(list(map(cls.dumps, obj)))}]"

        if isinstance(obj, dict):
            data = ",".join(
                [f'"{key}": {cls.dumps(value)}' for (key, value) in obj.items()]
            )
            return f"""{{{data}}}"""

    @classmethod
    def dump(cls, obj, fp: IO[str]) -> None:
        fp.write(cls.dumps(obj))

    @classmethod
    def loads(cls, s: str):
        res, _ = cls._loads(s, 0)
        return res

    @classmethod
    def load(cls, fp: IO[str]):
        return cls.loads(fp.read())

    @classmethod
    def _loads(
        cls, s: str, start_index: int
    ) -> tuple[bool | str | int | float | list | dict | None, int]:
        if s[start_index] == '"':
            return cls._loads_str(s, start_index)
        if s[start_index] == "[":
            return cls._loads_list(s, start_index)

        if s[start_index].isdigit():
            return cls._loads_num(s, start_index)

        if s[start_index] == "t":
            return True, start_index + 4

        if s[start_index] == "f":
            return False, start_index + 5

        if s[start_index] == "n":
            return None, start_index + 4

        if s[start_index] == "{":
            return cls._loads_dict(s, start_index)

    @classmethod
    def _loads_list(cls, s: str, start_index: int) -> tuple[list, int]:
        end_index = start_index + 1
        braces = 1
        while braces:
            if s[end_index] == "[":
                braces += 1
            if s[end_index] == "]":
                braces -= 1

            end_index += 1

        arr = []
        index = start_index + 1
        while index < end_index - 2:
            while s[index] in (" ", ",", "\n"):
                index += 1
            res, index = cls._loads(s, index)
            arr.append(res)

        return arr, end_index

    @classmethod
    def _loads_dict(cls, s: str, start_index: int) -> tuple[dict, int]:
        end_index = start_index + 1
        braces = 1
        while braces:
            if s[end_index] == "{":
                braces += 1
            if s[end_index] == "}":
                braces -= 1

            end_index += 1

        index = start_index + 1
        result = {}
        while index < end_index - 2:
            while s[index] in (" ", ",", "\n"):
                index += 1
            key, index = cls._loads_str(s, index)

            while s[index] in (" ", ",", "\n", ":"):
                index += 1

            value, index = cls._loads(s, index)
            result[key] = value

        return result, end_index + 1

    @classmethod
    def _loads_str(cls, s: str, start_index: int) -> tuple[str, int]:
        end_index = start_index + 1
        while s[end_index] != '"':
            end_index += 1

        return s[start_index + 1 : end_index], end_index + 1

    @classmethod
    def _loads_num(cls, s: str, start_index: int) -> tuple[int | float, int]:
        end_index = start_index + 1
        while len(s) > end_index and (s[end_index].isdigit() or s[end_index] == "."):
            end_index += 1

        num = s[start_index:end_index]

        if num.count("."):
            return float(num), end_index

        return int(num), end_index



from .custom_serializer import CustomSerializer
from ..encoder.encoder import Encoder
from ..serializers import JsonSerializer, XmlSerializer
from typing import Literal


class SerializerFactory:
    _serializers_map = {
        "json": JsonSerializer,
        "xml": XmlSerializer,
    }

    @classmethod
    def create_serializer(cls, serializer_type: Literal["json", "xml"]):
        serializer = cls._serializers_map.get(serializer_type)

        if serializer is None:
            raise Exception(f"No such serializer format: {serializer_type}")

        return CustomSerializer(Encoder, serializer())



from setuptools import setup

setup(
    name="aliaksandr-custom-serializer",
    version="0.1.0",
    packages=[
        "custom_serializer",
        "custom_serializer.encoder",
        "custom_serializer.factory",
        "custom_serializer.serializers",
    ],
    entry_points={
        "console_scripts": [
            "custom-serialize = custom_serializer.custom_serializer:main"
        ]
    },
    url="",
    license="",
    author="alejandbel",
    author_email="",
    description="",
)



from typing import IO

from .base_serializer import BaseSerializer


class XmlSerializer(BaseSerializer):
    @classmethod
    def dumps(cls, obj) -> str:
        if isinstance(obj, bool):
            return f"<bool>{str(obj)}</bool>"

        if isinstance(obj, int):
            return f"<int>{str(obj)}</int>"

        if isinstance(obj, float):
            return f"<float>{str(obj)}</float>"

        if isinstance(obj, str):
            return f"<str>{obj}</str>"

        if isinstance(obj, type(None)):
            return f"<none>None</none>"

        if isinstance(obj, (list, tuple)):
            return f"<list>{''.join(list(map(cls.dumps, obj)))}</list>"

        if isinstance(obj, dict):
            data = "".join(
                [f"<{key}>{cls.dumps(value)}</{key}>" for (key, value) in obj.items()]
            )
            return f"<dict>{data}</dict>"

    @classmethod
    def dump(cls, obj, fp: IO[str]) -> None:
        fp.write(cls.dumps(obj))

    @classmethod
    def loads(cls, s: str):
        res, _ = cls._loads(s, 0)
        return res

    @classmethod
    def load(cls, fp: IO[str]):
        return cls.loads(fp.read())

    @classmethod
    def _loads(
        cls, s: str, start_index: int
    ) -> tuple[bool | str | int | float | list | dict | None, int]:
        index = start_index

        if s[index] != "<":
            raise Exception(f"Invalid symbol at position {index}")

        type_start_index = index + 1
        type_end_index = index

        while s[type_end_index] != ">":
            type_end_index += 1

        object_type = s[type_start_index:type_end_index]
        method_name = f"_loads_{object_type}"

        if not hasattr(cls, method_name):
            raise Exception("Unknown type")

        index = type_end_index + 1
        return getattr(cls, method_name)(s, index)

    @classmethod
    def _loads_str(cls, s: str, start_index: int) -> tuple[str, int]:
        end_index = start_index
        while s[end_index : end_index + 6] != "</str>":
            end_index += 1

        return s[start_index:end_index], end_index + 6

    @classmethod
    def _loads_bool(cls, s: str, start_index: int) -> tuple[bool, int]:
        end_index = start_index
        while s[end_index : end_index + 7] != "</bool>":
            end_index += 1

        bool_obj = s[start_index:end_index]
        if bool_obj == "True":
            return True, end_index + 7
        else:
            return False, end_index + 7

    @classmethod
    def _loads_int(cls, s: str, start_index: int) -> tuple[int, int]:
        end_index = start_index
        while s[end_index : end_index + 6] != "</int>":
            end_index += 1

        int_obj = s[start_index:end_index]
        return int(int_obj), end_index + 6

    @classmethod
    def _loads_float(cls, s: str, start_index: int) -> tuple[float, int]:
        end_index = start_index
        while s[end_index : end_index + 8] != "</float>":
            end_index += 1

        int_obj = s[start_index:end_index]
        return float(int_obj), end_index + 8

    @classmethod
    def _loads_none(cls, s: str, start_index: int) -> tuple[type(None), int]:
        end_index = start_index
        while s[end_index : end_index + 7] != "</none>":
            end_index += 1

        return None, end_index + 7

    @classmethod
    def _loads_list(cls, s: str, start_index: int) -> tuple[list, int]:
        end_index = start_index
        deep = 1
        while deep:
            if s[end_index : end_index + 6] == "<list>":
                deep += 1
            if s[end_index : end_index + 7] == "</list>":
                deep -= 1

            end_index += 1

        end_index -= 1
        arr = []
        index = start_index
        while index < end_index:
            res, index = cls._loads(s, index)
            arr.append(res)

        return arr, end_index + 7

    @classmethod
    def _loads_dict(cls, s: str, start_index: int) -> tuple[dict, int]:
        end_index = start_index
        deep = 1
        while deep:
            if s[end_index : end_index + 6] == "<dict>":
                deep += 1
            if s[end_index : end_index + 7] == "</dict>":
                deep -= 1

            end_index += 1

        end_index -= 1

        index = start_index
        result = {}

        while index < end_index:
            ket_start_index = index + 1
            key_end_index = index + 1

            while s[key_end_index] != ">":
                key_end_index += 1

            key = s[ket_start_index:key_end_index]

            value, index = cls._loads(s, key_end_index + 1)
            index += 3 + len(key)

            result[key] = value

        return result, end_index + 7



import inspect


def is_iterable(obj):
    return (
        hasattr(obj, "__iter__")
        and hasattr(obj, "__next__")
        and callable(obj.__iter__)
        and obj.__iter__() is obj
    )


def get_class_that_defined_method(meth):
    cls = getattr(
        inspect.getmodule(meth),
        meth.__qualname__.split(".<locals>", 1)[0].rsplit(".", 1)[0],
        None,
    )
    if isinstance(cls, type):
        return cls
